//******************************************************************
//
//	MP3 player kit BK1 Boot Loader
//
//	MCU:ATmega168
//		efuse:0xF8	u[g̈ 1024word ZbgxN^̓u[g̈
//		hfuse:0xD5	`bv EEPROMی, BOD 2.7V
//		lfuse:0xE2	CR 8MHz
//
//	Compiler:	AVR-GCC		4.2.2
//				AVR-LIBC	1.4.7
//				binutils	2.18
//
//	2008.7.27	SPI Mode 0, SHDC̃nOAbv, ubNTCY 512ݒ
//	2007.3.23	KAWAKAMI Yukio
//
//	New BSD license
//******************************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/boot.h>
#include	<avr/eeprom.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"BK1boot.h"

#define		FILE_NAME	"MP3BK1  BIN"

uchar	DiskCache[512];	// fBXNLbV
ulong	DiskCacheAddr;

uchar	FATtype;			// 1:FAT12  2:FAT16
uchar	SectorsPerCluster;	// 1NX^̃ZN^
ushort	ReservedSectors;	// \ZN^
uchar	FATcount;			// FAT̐ ʏ 2
uint	RootDirEntriesCount;	// [gfBNgGg[
ulong	SectorCount;	// ZN^
ushort	FATsectors;		// FATggZN^
uint	FATstart;		// FATJnZN^
uint	DIRstart;		// [gfBNgJnZN^
ulong	DataStart;		// f[^̈JnZN^
//ushort	ClusterCount;	// NX^


//****************************************************************************
// JIvV -Ttext=0x3800 Ǝw肷̂
// ʂ̃R[h 0x3800(u[g[_̈)ɔzu
//****************************************************************************

//*******************************************************
//	C[`
//*******************************************************
int main (void) {
	struct DIRECTORY dir;
	struct FILE_DATA fd;

	cli();	// 荞݋֎~

	// |[gBgpś͓{vAbv
	PORTB = (1<<PB7)|(1<<PB6)|(1<<MMC_CS)|(1<<MMC_MOSI)|(1<<MMC_SCK)|(1<<MP3_CTRL_CS)|(1<<MP3_DATA_CS);
	DDRB = (1<<MMC_SCK)|(1<<MMC_CS)|(1<<MMC_MOSI)|(1<<MP3_CTRL_CS)|(1<<MP3_DATA_CS);

	PORTC = (1<<SCL)|(1<<SDA)|(1<<POWER);	// dێ
	DDRC = (1<<POWER)|(1<<MP3_RESET);

	PORTD = (1<<SW_PORT_START)|(1<<SW_PORT_REW)|(1<<SW_PORT_VOLUP)|(1<<SW_PORT_VOLDOWN)|(1<<ROM_CS)|(1<<TXD)|(1<<RXD);
	DDRD = (1<<LED);

	DiskCacheAddr = 0xFFFFFFFF;

	if ((mmc_init() == 0)&&		// MMC/SDJ[h
		(!read_VFAT_info())&&	// FAT12/16
		(!search_file(&dir))&&	// t@C
		(!check_file(&dir))){	// t@CL

		// t@CI[v
		open_file(&dir, &fd);
	
		// ZtvO~O
		spm_main(&fd);

		// ^CX^v
		eeprom_write_block(&dir.time, EEPROM_FIRM_TIMESTAMP, 4);
		eeprom_busy_wait();	// ݏI҂
	}

	boot();

	return 0;
}


//****************************************************************
// ZtvO~OEC
//****************************************************************
uchar spm_main(struct FILE_DATA *fd){
	uint	i;
	ushort	w;
	ushort	sum;
	ulong	page;

	// EEPROMݏI҂
//	read_eeprom(0);

	page = 0;
	sum = 0;

	while(page < fd->size){
		// MMC/SDPZN^ǂݍ
		if (mmc_read(fd->sec)){
			return 1;
		}

		PIND |= (1<<LED);	// LED_

		for (i = 0; i < 512; i += 2){
			if ((i & (SPM_PAGESIZE - 1)) == 0){
				// y[W
				boot_page_erase(page);
				boot_spm_busy_wait();
				sum = 0xFFFF;
			}
			w = *(ushort *)(&DiskCache[i]);
			sum &= w;
			boot_page_fill(page + i, w);
			if ((i & (SPM_PAGESIZE - 1)) == (SPM_PAGESIZE - 2)){
				if (sum != 0xFFFF){	// S FFFF Ȃ珑܂Ȃ
					// y[W
					boot_page_write(page);
					boot_spm_busy_wait();
				}
				page += SPM_PAGESIZE;
			}
		}

		next_sector(fd);
	}

	// RWWZNVLɂ
	boot_rww_enable();

	return 0;
}


//****************************************************************
//	OԒñvOs
//****************************************************************
ADRS3400H void boot(void){
	ushort	code;

	// R[h`FbNBŏ jmp߂
	code = pgm_read_word(0);
	if ((code & 0xFE0E) == 0x940C){

		// 0ԒnɃWv(RESET)
		asm volatile("jmp	0\n"::);
	}

	PORTC &= ~(1<<POWER);	// POWER OFF
}


//*******************************************************
//	ȒPȃt@C`FbN
//*******************************************************
ADRS3400H uchar check_file(struct DIRECTORY *dir){
	ulong	sec;
	ulong	timestamp;
	ulong	filetime;

	if (dir->size > (0x3400))	return 1;	// TCY 13KB傫

	if (pgm_read_byte(0) != 0xFF){
		// EEPROMɋL^ꂽt@C̃^CX^vr
		eeprom_read_block(&timestamp, EEPROM_FIRM_TIMESTAMP, 4);
		filetime = *((ulong *)&(dir->time));
		if (timestamp == filetime)	return 1;				// t@[
	}
	
	// ŏ̃ZN^ǂ
	sec = cluster_to_adrs(dir->cluster);
	if (mmc_read(sec))	return 2;				// ǂݍ݃G[

	// ŏ jmp߂(Flashe 16KBȏ AVȐꍇ jmp߂ɂȂ)
	if (((DiskCache[0] & 0x0E) != 0x0C)||
		((DiskCache[1] & 0xFE) != 0x94)){
		return 3;	// f[^G[
	}

	return 0;
}


//*********************************************************
//	[gfBNǧŒt@C̃t@CT[`
//*********************************************************
ADRS3400H uchar search_file(struct DIRECTORY *dir){
	ulong	adrs;
	uint	i;

	adrs = (ulong)DIRstart * 512;
	for (i = 0; i < RootDirEntriesCount; i++){
		read_32(adrs, (uchar *)dir);
		adrs += 32;
		if (dir->filename[0] == 0)	break;
		if ((!(dir->att & 0x18))&&
			(strncmp_P((char *)dir, PSTR(FILE_NAME), 11) == 0)){
			return 0;	// 
		}
	}
	return 1;	// 
}


//**********************************************
//	t@CI[v
//**********************************************
void open_file(struct DIRECTORY *dir, struct FILE_DATA *fd){

	fd->size = (ushort)dir->size;
	fd->sc = 0;
	fd->cluster = dir->cluster;
	fd->sec = cluster_to_adrs(fd->cluster);
	fd->p = 0;
}


//**********************************************
//	̃ZN^w
//**********************************************
void next_sector(struct FILE_DATA *fd){

	fd->sec += 512;
	fd->p += 512;
	if (++(fd->sc) >= SectorsPerCluster){
		fd->cluster = next_cluster(fd->cluster);
		fd->sec = cluster_to_adrs(fd->cluster);
		fd->sc = 0;
	}
}


//*********************************************************
//	NX^ԍf[^i[AhX擾
//*********************************************************
ulong cluster_to_adrs(uint cluster){

	return (((ulong)cluster - 2) * (ulong)SectorsPerCluster + DataStart) * 512;
}


//*********************************************************
//	̃NX^߂
//*********************************************************
uint next_cluster(uint c){
	ulong	addr;
	uint	data;
	ulong	fatadr;

	if (FATtype == FAT12){
		// FAT12
		addr = (ulong)FATstart * 512 + (ulong)c + (c >> 1);
		if ((addr & 511) == 511){
			// ZN^܂
			data = read_word(addr) & 0x00FF;
			data |= ((read_word(addr+1) & 0x00FF) << 8);
		} else {
			fatadr = addr & 0xFFFFFE00;
			data = read_word(addr);
		}
		if (c & 1){
			c = data >> 4;
		} else {
			c = data & 0x0fff;
		}
		if (c >= 0x0ff8)	c = 0xFFFF;
	} else {
		// FAT16
		addr = (ulong)FATstart * 512 + (ulong)c * 2;
		fatadr = addr & 0xFFFFFE00;
		c = read_word(addr);
		if (c >= 0xFFF8)	c = 0xFFFF;
	}

	return c;
}


//*******************************************************
//	MMC/SDJ[h 2byteǂ
//	ZN^ẼANZX͂łȂ̂
//	ʍHōl邱
//*******************************************************
uint read_word(ulong addr){
	uint	w;
	uint	b;

	b = addr & 511;
	addr &= 0xFFFFFE00;

	mmc_read(addr);

	w = (uint)DiskCache[b++];
	w |= ((uint)DiskCache[b] << 8);

	return w;
}


//*******************************************************
// MMCJ[hPZN^ǂݍ
//*******************************************************
uchar mmc_read(ulong addr){
	int		i;
	uchar	r;
	uchar	*buff;

	// fBXNLbVAhXr
	if (addr == DiskCacheAddr)	return 0;
	DiskCacheAddr = addr;
	
	MMC_CS_ON;
	r = SPI_command(17, addr);
	if (r == 0){
		for (i = 0; i < 30000; i++){	// ̂҂ꍇ
			r = SPI_in();
			if (r == 0xFE){
				break;		// ]Jn}[N
			}
		}
	}

	if (r != 0xFE){
		r = 1;
		goto QUIT;		// ERROR
	} else {
		r = 0;
	}

	buff = DiskCache;
	for (i = 0; i < 512; i++){
		*buff++ = SPI_in();
	}

	SPI_in();	// CRC
	SPI_in();	// CRC

  QUIT:
	MMC_CS_OFF;
	SPI_in();

	return r;
}


//****************************************
//	MMC/SD SPIR}ho
//****************************************
uchar SPI_command(uchar com, ulong arg){
	uint	i;
	uchar	r;

	// BUSY`FbN
	for (i = 0; i < 30000; i++){
		r = SPI_in();
		if (r == 0xFF)	break;
	}
	if (r != 0xFF)	return 1;

	SPI_out(com|0x40);
	SPI_out(arg >> 24);
	SPI_out(arg >> 16);
	SPI_out(arg >> 8);
	SPI_out(arg);
	SPI_out(0x95);	// R}hCRC

	// R1҂
	for (i = 0; i < 10000; i++){
		r = SPI_in();
		if (r != 0xFF)	break;
	}

	return r;
}


//****************************************
//	SPIf[^o
//****************************************
void SPI_out(uchar data){

	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
}


//****************************************
//	SPIf[^
//****************************************
uchar SPI_in(void){

	SPDR = 0xFF;
	while(!(SPSR & (1<<SPIF)));
	return SPDR;
}


//*******************************************************
//	MMC/SDJ[h 32byteǂ buffɊi[
//	DiskCachegƂɒ
//*******************************************************
uchar read_32(ulong addr, uchar *buff){
	uint	b;
	uchar	i;

	b = addr & 511;
	addr &= 0xFFFFFE00;
	
	if (mmc_read(addr))	return 1;

	for (i = 0; i < 32; i++){
		*buff++ = DiskCache[b++];
	}

	return 0;
}


//*********************************************************
//	MMC/SDJ[h̏
//*********************************************************
ADRS3400H uchar mmc_init(void){
	uchar	c;
	uint	i;

	// SPI}X^, Mode 0, Clock 8MHz/64=125kHz
	SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR0)|(1<<SPR1);
	SPSR = (1<<SPI2X);

	//MMC_CS_OFF;	// CS=H

	// MMC/SD SPI[hڍsÕNbNM
	for (c = 0; c < 255; c++){	// ԑ҂˂ĒĂ
		SPI_in();
	}

	// SPI[hڍsR}h
	MMC_CS_ON;	// CS=L
	c = SPI_command(0, 0);			// CMD0
	if (c == 0xFF)	return 0xFF;	// J[hȂ

	for (i = 0; i < 500; i++){
		c = SPI_command(1, 0);		// CMD1
		if ((c & 0x81) == 0)	break;
	}

	if (c == 0){
		SPI_command(16, 512);		// CMD16  ubNTCY 512bytes
	}
	
	MMC_CS_OFF;	// CS=H
	SPI_in();

	// SPI}X^, Mode 0, Clock 8MHz/2=4MHz
	SPCR = (1<<SPE)|(1<<MSTR);

	return c;
}


//*********************************************************
//	VFATǂݎ
//*********************************************************
ADRS3400H uchar read_VFAT_info(void){
	uchar	buff[32];
	ulong	sec;
	ushort	bpb;
	ushort	n;

	// VFATǂݎBZN^ 512bytesŌߑłĂ
	read_32(446, buff);								// MBRǂݍ
	bpb = *((ushort *)&buff[8]);					// BPB擪ZN^

	read_32((ulong)bpb * 512, buff);				// ǂݍ
	n = *((ushort *)&buff[11]);						// ZN^
	if (n != 512){
		return 2;	// 1ZN^ 512bytesłȂ
	}

	FATcount = buff[16];		// FAT̐ ʏ 2

	SectorsPerCluster = buff[13];		// NX^̃ZN^
	ReservedSectors = *((ushort *)&buff[14]);	// \ZN^
	FATstart = bpb + ReservedSectors;			// FATJnZN^

	RootDirEntriesCount = *((ushort *)&buff[17]);	// [gfBNgGg

	FATsectors = *((ushort *)&buff[22]);		// FAT̃ZN^
	if (FATsectors == 0){
		return 1;	// FAT32炵
	}

	DIRstart = (ushort)FATstart + (ushort)FATcount * FATsectors;	// FAT̐~ZN^
	n = RootDirEntriesCount / 16;	// [gfBNgɕKvȃZN^
//	if (RootDirEntriesCount % 16){
//		n++;	// Ô߁A]肪o؂グBʏ͕svƎv
//	}
	DataStart = (ulong)DIRstart + (ulong)n;		// f[^i[̈

	sec = (ulong)*((ushort *)&buff[19]);		// ZN^
	if (sec == 0){
		read_32((ulong)bpb * 512 + 32, buff);	// ǂݍ
		sec = *((ulong *)&buff[0]);
	}
	SectorCount = sec;
	sec -= FATstart;
	sec /= SectorsPerCluster;
	if (sec >= 0x1000){
		FATtype = FAT16;	// FAT16
	} else {
		FATtype = FAT12;	// FAT12
	}

	// ǂݍ݂Ȃȉ͕̏sv
	
	// f[^̈ZN^
//	sec = SectorCount - ReservedSectors - FATsectors * FATcount - (ulong)RootDirEntriesCount * 32 / 512;

	// NX^
//	ClusterCount = (ushort)(sec / SectorsPerCluster) + 2;

	return 0;
}

